@@ -1,6 +1,6 @@
 /*
     Clif - A C-like Interpreter Framework
-    Copyright (C) 1998 L. Koren
+    Copyright (C) 1998, 2000 L. Koren
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -21,6 +21,7 @@
 
 #define BACKSLASH '\\'
 #define SP ' '
+#define ALERT '\a'
 #define BS '\b'
 #define FF '\f'
 #define NL '\n'
@@ -32,6 +33,7 @@
 #define DOT '.'
 #define STAR '*'
 #define CIRCUMFLEX '^'
+#define TICK '\''
 
 extern int yyleng;
 
@@ -60,28 +62,36 @@ void
 s_conv (yytext)
      char *yytext;
 {
-  int i, j, help;
-  
+  int i, j, help, wide_string = 0;
+
+  if (yytext[0] == 'L')
+    {
+      yytext++;
+      wide_string = 1;
+    }
+
   if (proc)
     num_args[proc] = args[proc] + 1;
 
-  for (i = 1,j = 0; i < yyleng-1; i++,j++)
+  for (i = 1, j = 0; i < yyleng - 1; i++, j++)
     {
       if (BACKSLASH == yytext[i])
 	{
 	  switch (yytext[i + 1])
 	    {
-	    case 'b' : yytext[j] = BS; i++;
+	    case 'a': yytext[j] = ALERT; i++;
 	      break;
-	    case 'f' : yytext[j] = FF; i++;
+	    case 'b': yytext[j] = BS; i++;
 	      break;
-	    case 'n' : yytext[j] = NL; i++;
+	    case 'f': yytext[j] = FF; i++;
 	      break;
-	    case 'r' : yytext[j] = CR; i++;
+	    case 'n': yytext[j] = NL; i++;
 	      break;
-	    case 't' : yytext[j] = HT; i++;
+	    case 'r': yytext[j] = CR; i++;
 	      break;
-	    case 'v' : yytext[j] = VT; i++;
+	    case 't': yytext[j] = HT; i++;
+	      break;
+	    case 'v': yytext[j] = VT; i++;
 	      break;
 	      /* 
 	       * Control characters
@@ -99,6 +109,14 @@ s_conv (yytext)
 	    case 'x' : sscanf (&yytext[i + 2], "%2x", &help); i += 3;
 	      yytext[j] = help;
 	      break;
+	    case TICK:
+	      yytext[j] = TICK;
+	      i++;
+	      break;
+	    case '\"':
+	      yytext[j] = '\"';	/* " */
+	      i++;
+	      break;
 	    }
 	}
       else if ('%' == yytext[i] && '%' != yytext[i + 1])
@@ -110,6 +128,12 @@ s_conv (yytext)
 	      continue;
 	    }
 	  yytext[j++] = yytext[i++];
+
+	  if (wide_string == 1)
+	    {
+	      error_message (6018);
+	      wide_string++;
+	    }
 	  
 	  while (1)
 	    {
@@ -343,11 +367,114 @@ s_conv (yytext)
 	  yytext[j++] = yytext[i++];
 	  yytext[j] = yytext[i];
 	}
-      else
+      else if ('?' == yytext[i] && '?' == yytext[i + 1])
 	{
-	  yytext[j] = yytext[i];
+	  error_message (6032);
+	  /* Brain-damaged trihraphs handling. */
+	  switch (yytext[i + 2])
+	    {
+	    case '=':
+	      yytext[j] = HASH;
+	      i += 2;
+	      break;
+	    case '(':
+	      yytext[j] = '[';
+	      i += 2;
+	      break;
+	    case '/':
+	      yytext[j] = BACKSLASH;
+	      i += 2;
+	      if ('\0' != yytext[i + 1])
+		{
+		  register int k;
+		  /* We have to re-process the escape
+		     characters. First, copy string to the new
+		     location. (We have replaced the trigraph '??/'
+		     already.) Then, all control variables (i, j,
+		     yyleng) must be reset. */
+		  yytext[j + 1] = yytext[++i];
+		  for (k = j + 2, i++; '\0' != yytext[i]; k++, i++)
+		    yytext[k] = yytext[i];
+		  /* Should be null all the time. */
+		  yytext[k] = yytext[i];
+		  /* We have to change the yyleng, because the string
+                     shrunk. */
+		  yyleng = k;
+		  j--;
+		  i = j;
+		}
+	      break;
+	    case ')':
+	      yytext[j] = ']';
+	      i += 2;
+	      break;
+	    case TICK:
+	      yytext[j] = CIRCUMFLEX;
+	      i += 2;
+	      break;
+	    case '<':
+	      yytext[j] = '{';
+	      i += 2;
+	      break;
+	    case '!':
+	      yytext[j] = '|';
+	      i += 2;
+	      break;
+	    case '>':
+	      yytext[j] = '}';
+	      i += 2;
+	      break;
+	    case '-':
+	      yytext[j] = '~';
+	      i += 2;
+	      break;
+	    default:
+	      yytext[j] = yytext[i];
+	    }
 	}
+      else if ('"' == yytext[i])
+	{
+	  char *buf = NULL;
+
+	  while ('"' != yytext[++i])
+	    if ('L' == yytext[i] && '"' == yytext[i + 1])
+	      wide_string++;
+	    else if (' ' == yytext[i] || '\t' == yytext[i] ||
+		     '\v' == yytext[i] || '\b' == yytext[i] ||
+		     '\f' == yytext[i])
+	      ;
+	    else if ('\n' == yytext[i] || '\r' == yytext[i])
+	      {
+		extern int s;
+		extern FILEATTR spf[];
+		spf[s].line_counter++;
+	      }
+	    else if ('\0' == yytext[i])
+	      break;
+	    else
+	      {
+		buf = yytext + i;
+		break;
+	      }
+	  
+	  if (buf)
+	    {
+	      while (' ' != yytext[i] && '\t' != yytext[i] &&
+		     '\n' != yytext[i] && '\r' != yytext[i] &&
+		     '\f' != yytext[i] && '\v' != yytext[i] &&
+		     '\0' != yytext[i] && '\"' != yytext[i])
+		i++;
+	      yytext[i] = '\0';
+	      error_message (1004, buf);
+	    }
+	  j--;
+	}
+      else
+	yytext[j] = yytext[i];
     }
+	  
+  if (proc && wide_string)
+    error_message (6023);
   yytext[j] = '\0';
 }
 
@@ -381,9 +508,20 @@ compare_format_args (type1, type2, type)
   struct internal_type *type1, *type2;
   enum intern_arit_class type;
 {
-  if (NULL == type1 || NULL == type2)
+  if (NULL == type1 && NULL == type2)
     return;
-
+  else if (NULL == type1 && NULL != type2)
+    {
+      if (! ARRAY_P(type2) || ! CHAR_P(type2->output))
+	error_message (6022);
+      return;
+    }
+  else if (NULL != type1 && NULL == type2)
+    {
+      error_message (6021);
+      return;
+    }
+  
   if (call_by_value && SCANF_P && POINTER_P(type2))
     {
       type2 = type2->output;
@@ -396,7 +534,9 @@ compare_format_args (type1, type2, type)
       (ARRAY_P(type1) && ARRAY_P(type2)))
     {
       if (! VOID_P(type1->output) &&
-	  ! TYPES_EQ_P(type1->output, type2->output))
+	  ! TYPES_EQ_P(type1->output, type2->output) &&
+	  (! (CHAR_P(type1->output) && INTEGER_P(type2->output)) ||
+	   (INTEGER_P(type1->output) && CHAR_P(type2->output))))
 	error_message (6017, type1, args[proc]);
     }
   else if (DOUBLE_P(type1) && (type & FLT))
