@@ -173,6 +173,10 @@ static int allocate_aggregate PROTO((struct internal_type *));
 /* Static variables are put into tables. */
 static void putstruct_static PROTO((struct ident_tab_loc *));
 
+static int memory_size PROTO((int *, struct internal_type *));
+static void check_init_bracket PROTO((int *, struct internal_type *,
+				      char *));
+
 int
 hastab_init ()
 {
@@ -521,6 +525,7 @@ has (var_name)
 	{
 	  /* There is no list. */
 	  np->name = var_name;
+	  np->declaration_line = spf[s].line_counter;
 	  putstruct (np);
 	}
       else
@@ -537,6 +542,7 @@ has (var_name)
 	  np = (struct tab *) allocate (sizeof (struct tab), PERM);
 	  arch->next = np;
 	  np->name = var_name;
+	  np->declaration_line = spf[s].line_counter;
 	  putstruct (np);
 	}
       pi++;
@@ -583,6 +589,14 @@ putstruct (np)
   identtab[pi].adr = NULL;
   identtab[pi].type = typeh[type_spec_count];
 
+  /* If it is a typedef, just put it in the table, do not allocate. */
+  if (TYPEDEF_P(typeh[type_spec_count]))
+    {
+      if (ARRAY_P(typeh[type_spec_count]->output))
+	put_array_subscript (&(identtab[pi].type->output));
+      return;
+    }
+  
   switch (typeh[type_spec_count]->attribute.function_class)
     {
     case LOCAL:
@@ -745,7 +759,13 @@ point_loc (var_name)
 {
   struct tab     *pa = NULL;
   struct ident_tab_header *walk = identtab_loc;
-  extern int offset[];
+  struct var_s 
+  {
+    char *adr;
+    int offset;
+    char *name;
+  };
+  extern struct var_s variable[];
   register int count = 0;
   
   if (identtab_loc)
@@ -772,7 +792,7 @@ point_loc (var_name)
 	}
       pa->count++;
       pa->use_line_number = spf[s].line_counter;
-      offset[++set] = count;
+      variable[++set].offset = count;
       return walk->table + pa->def;
     }
   else
@@ -805,6 +825,7 @@ has_loc (type, var_name)
 	{
 	  /* There is no list. */
 	  np->name = var_name;
+	  np->declaration_line = spf[s].line_counter;
 	  putstruct_loc (np, type);
 	}
       else
@@ -821,6 +842,7 @@ has_loc (type, var_name)
 	  np = (struct tab *) allocate (sizeof (struct tab), BLOCK);
 	  arch->next = np;
 	  np->name = var_name;
+	  np->declaration_line = spf[s].line_counter;
 	  putstruct_loc (np, type);
 	}
       identtab_loc->pi_loc++;
@@ -856,6 +878,10 @@ putstruct_loc (np, type)
   identtab_loc->table[identtab_loc->pi_loc].type =
     typeh[type_spec_count];
 
+  /* If it is a typedef, just put it in the table, do not allocate. */
+  if (TYPEDEF_P(typeh[type_spec_count]))
+    return;
+
   if (STATIC_P(typeh[type_spec_count]) && type == VAR)
     putstruct_static (&(identtab_loc->table[identtab_loc->pi_loc]));
 
@@ -1247,44 +1273,14 @@ link_function (i)
 
 
 void
-set_value (text)
-  char *text;
+set_value (var_name)
+  char *var_name;
 {
   struct tab     *pa;
   struct ident_tab_header *walk = identtab_loc;
-  char *beg, *end, *tmp_line, *var_name;
-  int n;
-  extern char *string PROTO((char *));
-
+      
   if (identtab_loc)
     {
-      n = strlen (text);
-      tmp_line = mallocx (n+1);
-      if (NULL == tmp_line)
-	{
-	  perror ("");
-	  abort ();
-	}
-      strcpy (tmp_line, text);
-      end = strrchr (tmp_line, '=');
-      if (NULL != end)
-	*end = '\0';
-      beg = strrchr (tmp_line, '=');
-      if (NULL == beg)
-	beg = strrchr (tmp_line, '(');
-      if (NULL == beg)
-	beg = strrchr (tmp_line, ';');
-      if (NULL == beg)
-	beg = tmp_line;
-      
-      for (; ! isalpha ((int) *beg); beg++);
-      for (end = beg; *end != ' ' && *end != '.' && *end != '-' &&
-	     *end != '[' && *end != '(' && *end != '\0'; end++);
-      *end = '\0';
-      
-      var_name = string (beg);
-      free (tmp_line);
-      
       while  (NULL == (pa = lookup_loc (var_name, walk)))
 	{
 	  if (walk->previous_level)
@@ -1761,7 +1757,14 @@ add_spec_to_tag ()
       error_message (1000);
       return -1;
     }
-  else if (NULL == tagtab->table[pa->def].type->input)
+  else if (TYPEDEF_P(tagtab->table[pa->def].type)
+	   && NULL == tagtab->table[pa->def].type->input)
+    {
+      tagtab->table[pa->def].type->output->input = l_type_spec;
+      l_type_spec = NULL;
+    }
+  else if (! TYPEDEF_P(tagtab->table[pa->def].type)
+	   && NULL == tagtab->table[pa->def].type->input)
     {
       tagtab->table[pa->def].type->input = l_type_spec;
       l_type_spec = NULL;
@@ -2406,6 +2409,13 @@ allocate_aggregate (type)
   int size = 0;
   int first_align =0;
 
+  if (TYPEDEF_P(type) && ! type->attribute.memory_size)
+    {
+      type->attribute.memory_size
+	= allocate_aggregate (type->output);
+      return 1;
+    }
+
   if (SIMPLE_P(type) && ! type->attribute.memory_size)
     {
       type->attribute.memory_size = allocate_struct (type);
@@ -2424,18 +2434,24 @@ allocate_aggregate (type)
   switch (type->attribute.function_class)
     {
     case ARRAY:
+    {
       if (!type->output->attribute.memory_size)
 	allocate_aggregate (type->output);
       size = 1;
+      if (0 >= ((struct range *)walk->attribute.domain)->upper)
+	return 0;
       while (walk)
 	{
 	  size *= (((struct range *) walk->attribute.domain)->upper
 	    - ((struct range *) walk->attribute.domain)->lower);
+	  walk->attribute.memory_size =
+	    size * type->output->attribute.memory_size;
 	  walk = walk->arity;
 	}
       type->attribute.memory_size = size
 	* type->output->attribute.memory_size;
-      break;
+    }
+    break;
     case STRUCT_FC:
       while (walk)
 	{
@@ -2615,3 +2631,226 @@ exit_file_scope ()
   if (identtab_loc && identtab_loc->file_scope == spf[s].name)
     identtab_loc->file_scope = NULL;
 }
+
+
+/* If the name is declared as typedef return it as TYPENAME token -
+   true (1); IDENT - false (0), otherwise. */
+int
+typedef_p (name)
+  char *name;
+{
+  struct tab *pa = NULL;
+  struct ident_tab_header *walk = identtab_loc;
+  
+  if (! typedef_f)
+    return 0;
+  
+  if (identtab_loc)
+    {
+      /* Walk through all tables of locals. */
+      while (NULL == (pa = lookup_loc (name, walk)))
+	{
+	  if (walk->previous_level)
+	    walk = walk->previous_level;
+	  else
+	    break;
+	}
+    }
+  if (pa && TYPEDEF_P((walk->table + pa->def)->type))
+    return 1;
+  else if (NULL == pa)
+    {
+      pa = lookup (name);
+      if (pa && TYPEDEF_P((identtab + pa->def)->type))
+	return 1;
+    }
+  return 0;
+}
+
+/* Returns the declaration line of a variable. Called from
+   error_message function. */
+int
+get_declaration_line (name)
+  char *name;
+{
+  struct tab *pa = NULL;
+  struct ident_tab_header *walk = identtab_loc;
+  
+  if (identtab_loc)
+    {
+      /* Walk through all tables of locals. */
+      while (NULL == (pa = lookup_loc (name, walk)))
+	{
+	  if (walk->previous_level)
+	    walk = walk->previous_level;
+	  else
+	    break;
+	}
+    }
+  if (pa)
+    return pa->declaration_line;
+
+  pa = lookup (name);
+  if (pa)
+    return pa->declaration_line;
+  return 0;
+}
+
+
+/* Walk recursively through the type tree. Return the size of the
+   subtype. */
+static int
+memory_size (level, walk)
+  int *level;
+  struct internal_type *walk;
+{
+  int result = 0;
+  
+  if (! *level && walk->attribute.memory_size)
+    return walk->attribute.memory_size;
+
+  if (*level && ! result && walk->output)
+    {
+      *level -= 1;
+      result = memory_size (level, walk->output);
+      if (ARRAY_P(walk) &&
+	  ! (STRUCT_P(walk->output) || UNION_P(walk->output)))
+	{
+	  /* If the array has as fields structures or unions, we must
+	     walk down the subtypes to know the sizes. */
+	  *level += 1;
+	  result = 0;
+	}
+    }
+
+  if (*level && ! result && walk->input)
+    {
+      if (! ARRAY_P(walk))
+	*level -= 1;
+      result = memory_size (level, walk->input);
+    }
+  
+  if (*level && ! result && walk->arity)
+    {
+      *level -= 1;
+      result = memory_size (level, walk->arity);
+    }
+
+  return result;
+}
+
+
+/* Find out if the number of brackets is equal to the number of
+   subtypes declared in the type, i.e. the initialization expression
+   is fully bracketed. */
+static void
+check_init_bracket (level, walk, buf)
+  int *level;
+  struct internal_type *walk;
+  char *buf;
+{
+  /* Trying to reconstruct the initializer in the case of an error. */
+  if (STRUCT_P(walk) && 
+      (strlen (walk->input->field_name) + strlen (buf) < BUFSIZ - 1))
+    {
+      strcat (buf, ".");
+      strcat (buf, walk->input->field_name);
+    }
+  else if ((ARRAY_P(walk) || walk->arity) && (strlen (buf) + 4 < BUFSIZ))
+    strcat (buf, "[0]");
+
+  if (walk->output)
+    {
+      *level += 1;
+      check_init_bracket (level, walk->output, buf);
+      if (ARRAY_P(walk))
+	{
+	  int i = strlen (buf) - 1;
+	  if (']' == buf[i])
+	    buf[i - 2] = '\0';
+	  *level -= 1;
+	}
+    }
+  if (walk->input)
+    {
+      *level += 1;
+      check_init_bracket (level, walk->input, buf);
+      if (walk->arity)
+	*level -= 1;
+    }
+  if (walk->arity)
+    {
+      *level += 1;
+      check_init_bracket (level, walk->arity, buf);
+    }
+}
+
+/* Wrapper to error reporting and memory size. */
+int
+get_memory_size (level, walk, initializer_number)
+  int level;
+  struct internal_type *walk;
+  int initializer_number;
+{
+  char buf[BUFSIZ];
+  static int aggregate_level;
+  int j;
+  
+  if (1 == initializer_number)
+    {
+      aggregate_level = 0;
+      if (strncpy (buf, text, BUFSIZ))
+	check_init_bracket (&aggregate_level, walk, buf);
+      if (aggregate_level > level)
+	error_message (6030, buf);
+    }
+
+  if (1 == level)
+    j = 1;
+  else
+    j = aggregate_level - level;
+  return memory_size (&j, walk);
+}
+
+
+/* Recursive funtion. Returns the size of the field in struct or
+   union. Called during variable initialization. */
+int
+get_field_size (pos, walk)
+  int *pos;
+  struct internal_type *walk;
+{
+  int result = 0;
+  
+  if (SIMPLE_P(walk) && walk->attribute.memory_size)
+    {
+      result =  walk->attribute.memory_size;
+      (*pos)--;
+    }
+
+  if (*pos && !SIMPLE_P(walk) && walk->output)
+    result = get_field_size (pos, walk->output);
+
+  if (*pos && (STRUCT_P(walk) || UNION_P(walk)) && walk->input)
+    result = get_field_size (pos, walk->input);
+  
+  return result;
+}
+
+
+/* If the local variable was not initialized, usage counter is reset. */
+void
+noninitialized_loc (var_name)
+  char *var_name;
+{
+  struct tab     *pa = NULL;
+  
+  if (identtab_loc)
+    pa = lookup_loc (var_name, identtab_loc);
+
+  if (pa)
+    {
+      pa->count--;
+      pa->use_line_number = 0;
+    }
+}
