@@ -7,16 +7,18 @@ typedef union	{
   long double myldouble;
   float myfloat;
   char *mystring;
+  wchar_t *mywstring;
   char mychar;
 } YYSTYPE;
-#define	NUMBERI	258
-#define	NUMBERUI	259
-#define	NUMBERLI	260
-#define	NUMBERLUI	261
-#define	NUMBERD	262
-#define	NUMBERLD	263
-#define	NUMBERF	264
-#define	STRINGC	265
+#define	NUMBERI	257
+#define	NUMBERUI	258
+#define	NUMBERLI	259
+#define	NUMBERLUI	260
+#define	NUMBERD	261
+#define	NUMBERLD	262
+#define	NUMBERF	263
+#define	STRINGC	264
+#define	WSTRINGC	265
 #define	NUMBERC	266
 #define	AUTO	267
 #define	STATIC	268
